/** 
    @file exabayes.cpp
    
    @brief This file sets the flavour of exabayes that has been
    compiled (i.e., sequential, pll, examl, ...)
    
*/ 



// TODO re-activate all that initial bla bla when starting up
// exa-bayes (model / program info )


#ifdef HAVE_AVX
#define __AVX
#endif

#include <iostream>
#include <sstream>
#include <chrono>

#include "releaseDate.hpp" 

int NUM_BRANCHES; 


#define _INCLUDE_DEFINITIONS
#include "GlobalVariables.hpp"
#undef _INCLUDE_DEFINITIONS

#include "time.hpp"

#include "config/CommandLine.hpp"
#include "SampleMaster.hpp"
#include "comm/ParallelSetup.hpp"

#include "teestream.hpp"


#include "axml.h" 

// #define TEST  
#ifdef TEST
#endif

// have ae look at that later again 
double fastPow(double a, double b)
{
  union 
  {
    double d;
    int x[2];
  } u = { a };
  u.x[1] = (int)(b * (u.x[1] - 1072632447) + 1072632447);
  u.x[0] = 0;
  return u.d;
}

/**
   @brief the main ExaBayes function.

   @param tr -- a tree structure that has been initialize in one of the adapter mains. 
   @param adef -- the legacy adef
 */
static void exa_main ( CommandLine &cl,  std::shared_ptr<ParallelSetup> pl )
{   
  timeIncrement = CLOCK::system_clock::now(); 


#ifdef TEST     
  
  // double t = pow(2,128); 
  // printf("%f\n", t); 


  // double t2 = 1; 
  // for(nat i = 0; i < 128; ++i)
  //   t2 *= 2; 
  // printf("%f\n", )


  // auto mesh = IncompleteMesh{} ;
  // mesh.initialize(16, 1, 4);	

  // for(nat i = 0; i < 16 ; ++i)
  //   {
  //     auto coords = mesh.getCoordinates(i);
  //     auto rank = mesh.getRankFromCoordinates(coords); 
  //     tout <<  i << " has coordinates "  << coords[0] << "," << coords[1] << "," << coords[2] << "\t" << rank << std::endl; 
      
  //   }
  
  // ParallelSetup::genericExit(0); 
#else 
  // assert(0); 
  auto&& master = SampleMaster(   );
  
  master.setParallelSetup(pl); 
  master.setCommandLine(cl); 


  master.initializeRuns(Randomness(cl.getSeed())); 
  if( cl.isDryRun())
    {
      std::cout << "Command line, input data and config file is okay. Exiting gracefully." << std::endl; 
      ParallelSetup::genericExit(0); 
    }
  else 
    {
      master.run();
      master.finalizeRuns();
    }
#endif
}


/* 
   tell the CPU to ignore exceptions generated by denormalized floating point values.
   If this is not done, depending on the input data, the likelihood functions can exhibit 
   substantial run-time differences for vectors of equal length.
*/

static void ignoreExceptionsDenormFloat()
{
#if ! (defined(__ppc) || defined(__powerpc__) || defined(PPC))
  _mm_setcsr( _mm_getcsr() | _MM_FLUSH_ZERO_ON);
#endif   
}


static bool fileExists(const std::string &name)
{
  auto &&ifh = std::ifstream{name};
  bool result = ifh.is_open(); 
  ifh.close();
  return result; 
}


void makeInfoFile(const CommandLine &cl, const ParallelSetup &pl )
{
  auto &&ss = stringstream{}; 

  if(cl.isDryRun())
    ss << "/dev/null" ; 
  else 
    ss << OutputFile::getFileBaseName(cl.getWorkdir()) << "_info."  << cl.getRunid() ;

  if( not cl.isDryRun() && pl.isGlobalMaster() )
    {
      if(fileExists(ss.str()))
	{
	  std::cerr << pl << std::endl; 
	  std::cerr << std::endl <<  "File " << ss.str() << " already exists (probably \n"
		    << "from previous run). Please choose a new run-id or remove previous output files. " << std::endl; 
	  ParallelSetup::genericExit(-1); 
	}
    }

  globals.logFile = ss.str();   
  
  if( pl.isGlobalMaster())
    globals.logStream =  new std::ofstream(globals.logFile); 
  else 
    globals.logStream =  new std::ofstream{"/dev/null"}; 
    
  globals.teeOut =  new teestream(cout, *globals.logStream);

  if(not pl.isGlobalMaster())
    tout.disable(); 
}


static void initializeProfiler(const ParallelSetup& pl)
{
  // see this page for info 
  // http://google-perftools.googlecode.com/svn/trunk/doc/cpuprofile.html  
  // that option is important
  // CPUPROFILE_FREQUENCY=x
#ifdef _USE_GOOGLE_PROFILER

#if HAVE_PLL != 0 
  auto myProfileFile = "profile.out"; 
  remove(myProfileFile);
  ProfilerStart(myProfileFile);
#else  
  auto&& ss = std::stringstream {}; 
  ss << "profile.out." << pl.getGlobalRank() ; 
  auto myProfileFile = ss.str(); 
  remove(myProfileFile.c_str()); 
  ProfilerStart(myProfileFile.c_str()); 
#endif

#endif
}

 
void finalizeProfiler()
{
#ifdef _USE_GOOGLE_PROFILER
  ProfilerStop();
#endif
}


static void printInfoHeader(int argc, char **argv)
{  
  tout << endl; 

  tout << "This is "
#if HAVE_PLL != 0
       << "Yggdrasil, the sequential variant of "
#endif
       << PROGRAM_NAME
       << " version " << VERSION
       << ", a tool for Bayesian MCMC sampling of phylogenetic trees," 
#if HAVE_PLL == 0 
       << "\nbuild with the ExaML code base and MPI-support."
#else 
       << "\nbuild with the (Phylogenetic Likelihood Library) PLL code base for sequential execution."
#endif

       << std::endl <<  "This software has been released in " << RELEASE_DATE <<  " by \n\n\tAndre J. Aberer, Kassian Kobert and Alexandros Stamatakis" << std::endl 
       << "\nPlease send any bug reports, feature requests and inquiries to " << PACKAGE_BUGREPORT
       << std::endl << std::endl; 

  tout << PROGRAM_NAME << " was called as follows: " << endl; 
  for(int i = 0; i < argc; ++i)
    tout << argv[i] << " " ; 
  tout << endl 
       << "================================================================"
       << endl; 
}



// just having this, because of mpi_finalize
static int innerMain(int argc, char **argv)
{ 
  auto plPtr = make_shared<ParallelSetup>(); 

#if HAVE_PLL != 0 && ( (defined(_FINE_GRAIN_MPI) || defined(_USE_PTHREADS)))
  assert(0); 
#endif

  auto cl = CommandLine(); 
  cl.initialize(argc, argv); 

#if HAVE_PLL == 0 
  plPtr->initializeExaml(cl);
#endif

  ignoreExceptionsDenormFloat(); 

  initializeProfiler(*plPtr);

  makeInfoFile(cl, *plPtr);

  printInfoHeader(argc,argv); 
  
  exa_main( cl, plPtr); 
  
  finalizeProfiler();

  delete globals.logStream; 
  delete globals.teeOut;
  return 0;
}


int main(int argc, char **argv)
{ 
  ParallelSetup::initialize(argc, argv);
  innerMain(argc, argv); 
  ParallelSetup::finalize(); 
  return 0 ; 
}
